#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#define STACK_SIZE 10

/* Prototypes */
void *producer_main(void *);
void *consumer_main(void *);
void push(char e);
char pop();

/* Global variables implementing the stack */
char shared_stack[STACK_SIZE];
int top_element_index;

int main(int argc, char **argv) 
{
  int num_prod, num_cons;
  int num_iterations;
  int i;
  pthread_t *producers;
  pthread_t *consumers;

  /* Parse command-line arguments */
  if (argc != 4) {
    fprintf(stderr,"Usage: %s <#producers> <#consumers> <#iterations>\n",argv[0]);
    fprintf(stderr,"\t<#producers>: number of producers\n");
    fprintf(stderr,"\t<#consumers>: number of consumers\n");
    fprintf(stderr,"\t<#iterations>: number of times each producer produces\n");
    fprintf(stderr,"\tExample: %s 3 10 400\n",argv[0]);
    exit(1);
  }
  
  /* Check command-line arguments */
  if ((sscanf(argv[1],"%d",&num_prod) != 1) || 
      (sscanf(argv[2],"%d",&num_cons) != 1) || 
      (sscanf(argv[3],"%d",&num_iterations) != 1)) {
    fprintf(stderr,"Invalid command-line arguments\n");
    exit(1);
  }
  if ((num_prod < 1) || (num_cons < 1) || (num_iterations < 1)) {
    fprintf(stderr,"Invalid command-line arguments\n");
    exit(1);
  }
  if (num_prod * num_iterations % num_cons) {
    fprintf(stderr,"The number of producers times the number of iterations "
                   "must be divisible by the number of consumers\n");
    exit(1);
  }



  /* Initialize the stack to be empty */
  top_element_index = -1;	
  for (i=0; i < STACK_SIZE; i++) {
    shared_stack[i] = -1; 
  }

  /* Start all the threads */
  producers = (pthread_t *)calloc(num_prod, sizeof(pthread_t));
  for (i=0; i < num_prod; i++) {
    if (pthread_create(&(producers[i]),NULL,producer_main,
               (void*)num_iterations)) {
      fprintf(stderr,"Error while calling pthread_create()\n");
      exit(1);
    }
  }
  consumers = (pthread_t *)calloc(num_cons, sizeof(pthread_t));
  for (i=0; i < num_cons; i++) {
    if (pthread_create(&(consumers[i]),NULL,consumer_main,
               (void*)(num_iterations * num_prod / num_cons))) {
      fprintf(stderr,"Error while calling pthread_create()\n");
      exit(1);
    }
  }

  /* Wait for all the threads */
  for (i=0; i < num_prod; i++) {
    if (pthread_join(producers[i],NULL)) {
      fprintf(stderr,"Error while calling pthread_join()\n");
      exit(1);
    }
  }
  for (i=0; i < num_cons; i++) {
    if (pthread_join(consumers[i],NULL)) {
      fprintf(stderr,"Error while calling pthread_join()\n");
      exit(1);
    }
  }
  exit(0);
}

/* producer_main() */
void *producer_main(void *arg)
{
  int num_iterations = (int)arg;
  int i;


  for (i=0; i < num_iterations; i++) {
    char element = 65 + rand() % 26;
    push(element);
  }
  return NULL;
}

/* consumer_main() */
void *consumer_main(void *arg)
{
  int num_iterations = (int)arg;
  int i;

  for (i=0; i < num_iterations; i++) {
    char element = pop();
    element = 0;
  }
  return NULL;
}

/* push() */
void push(char e)
{
  // Busy wait
  while (top_element_index >= STACK_SIZE-1);

  // Add the element
  top_element_index++;

  // Print '!' if we can detect a  problem
  if (top_element_index > STACK_SIZE-1) fprintf(stderr,"!");
  if (shared_stack[top_element_index] != -1) fprintf(stderr,"!");

  // Update the stack
  shared_stack[top_element_index] = e;
  return;
}

/* pop() */
char pop()
{
  char element;

  // Busy wait
  while (top_element_index < 0);

  // Remove the element
  top_element_index--;
  element = shared_stack[top_element_index+1];
  shared_stack[top_element_index+1] = -1;

  // Print '!' if we can detect a  problem
  if (top_element_index < -1) fprintf(stderr,"!");
  if (element == -1) fprintf(stderr,"!");
 
  // Return the (former) top of the stack
  return element;
}
